import typing
from typing import Optional, TypeVar, Type
from types import TracebackType
import yarl
import asyncio
import aiohttp
import sys

from .gist import Gist
from .exceptions import AuthorizationFailure, NotFound
from .constants import API_URL


__all__ = ("Client",)


class Client:
    def __init__(self, access_token: str):
        self.access_token = access_token

        self._request_lock = asyncio.Lock()
        self.user_agent = f"Gists.py (https://github.com/witherredaway/gists.py) Python/{sys.version_info[0]}.{sys.version_info[1]} aiohttp/{aiohttp.__version__}"

    async def request(
        self,
        method: str,
        url: str,
        *,
        params=None,
        data=None,
        headers=None,
        authorization: bool = True,
    ) -> typing.Dict:
        """The method to make asynchronous requests to the GitHub API"""

        hdrs = {
            "Accept": "application/vnd.github.v3+json",
            "User-Agent": self.user_agent,
        }
        if authorization:
            hdrs["Authorization"] = "token %s" % self.access_token

        request_url = yarl.URL(API_URL) / url

        if headers is not None and isinstance(headers, dict):
            hdrs.update(headers)

        await self._request_lock.acquire()
        try:
            async with aiohttp.ClientSession() as session:
                response = await session.request(
                    method, request_url, params=params, json=data, headers=hdrs
                )
                remaining = response.headers.get("X-Ratelimit-Remaining")
                try:
                    data = await response.json()
                except aiohttp.client_exceptions.ContentTypeError:
                    data = response.content
                if response.status == 429 or remaining == "0":
                    reset_after = float(response.headers.get("X-Ratelimit-Reset-After"))
                    await asyncio.sleep(reset_after)
                    self._request_lock.release()
                    return await self.request(
                        method, request_url, params=params, data=data, headers=headers
                    )
                elif 300 > response.status >= 200:
                    return data
                elif response.status == 404:
                    raise NotFound(response, data)
                elif response.status == 401:
                    raise AuthorizationFailure(
                        "Invalid personal access token has been passed."
                    )

        finally:
            if self._request_lock.locked():
                self._request_lock.release()

    async def get_user_data(self) -> typing.Dict:
        user_data = await self.fetch_user()
        return user_data

    async def fetch_user(self) -> typing.Dict:
        """Fetch data of the authenticated user"""

        user_data: typing.Dict = await self.request("GET", "user")
        return user_data

    async def fetch_data(self, gist_id: str) -> typing.Dict:
        """Fetch data of a Gist"""

        gist_data: typing.Dict = await self.request(
            "GET", "gists/%s" % gist_id, authorization=False
        )
        return gist_data

    async def get_gist(self, gist_id: str) -> Gist:
        """Get a Gist object representing the gist associated with the provided gist_id"""

        data = await self.fetch_data(gist_id)
        return Gist(data, self)

    async def create_gist(
        self,
        files: typing.Dict,  # e.g. {"output.txt": {"content": "Content of the file"}}
        *,
        description: str = None,
        public: bool = True,
    ) -> Gist:
        """Create a new gist and return a Gist object associated with that gist"""

        data = {"public": public, "files": files}
        params = {"scope": "gist"}

        if description:
            data["description"] = description

        js = await self.request("POST", "gists", data=data, params=params)
        return Gist(js, self)

    async def update_gist(self, gist_id: str):
        """Re-fetch data and update the provided Gist object."""

        updated_gist_data = await self.fetch_data(gist_id)
        return updated_gist_data

    async def edit_gist(
        self,
        gist_id: str,
        *,
        files: typing.Dict,  # e.g. {"output.txt": {"content": "Content of the file"}}
        description: str = None,
    ) -> typing.Dict:
        """Edit the gist associated with the provided gist id, and return the edited data"""

        data = {"files": files}

        if description:
            data["description"] = description

        edited_gist_data = await self.request("PATCH", "gists/%s" % gist_id, data=data)
        return edited_gist_data

    async def delete_gist(self, gist_id: str):
        """Delete the gist associated with the provided gist id"""

        await self.request("DELETE", "gists/%s" % gist_id)



import typing
from typing import Optional, Union

import aiohttp

__all__ = (
    "ClientException",
    "HTTPException",
    "AuthorizationFailure",
    "NotFound",
)


class ClientException(Exception):
    """Exception that is raised when an operation in the Client class fails"""

    pass


# Taken from discord.py
def _flatten_error_dict(
    d: typing.Dict[str, typing.Any], key: str = ""
) -> typing.Dict[str, str]:
    items: typing.List[typing.Tuple[str, str]] = []
    for k, v in d.items():
        new_key = key + "." + k if key else k

        if isinstance(v, dict):
            try:
                _errors: typing.List[typing.Dict[str, typing.Any]] = v["_errors"]
            except KeyError:
                items.extend(_flatten_error_dict(v, new_key).items())
            else:
                items.append((new_key, " ".join(x.get("message", "") for x in _errors)))
        else:
            items.append((new_key, v))

    return dict(items)


# Taken from discord.py
class HTTPException(ClientException):
    """Exception that's raised when an HTTP request operation fails.

    Attributes
    ------------
    response: :class:`aiohttp.ClientResponse`
        The response of the failed HTTP request. This is an
        instance of :class:`aiohttp.ClientResponse`. In some cases
        this could also be a :class:`requests.Response`.
    text: :class:`str`
        The text of the error. Could be an empty string.
    status: :class:`int`
        The status code of the HTTP request.
    code: :class:`int`
        The Discord specific error code for the failure.
    """

    def __init__(
        self,
        response: aiohttp.ClientResponse,
        message: Optional[Union[str, typing.Dict[str, typing.Any]]],
    ):
        self.response: aiohttp.ClientResponse = response
        self.status: int = response.status  # type: ignore # This attribute is filled by the library even if using requests
        self.code: int
        self.text: str
        if isinstance(message, dict):
            self.code = message.get("code", 0)
            base = message.get("message", "")
            errors = message.get("errors")
            if errors:
                errors = _flatten_error_dict(errors)
                helpful = "\n".join("In %s: %s" % t for t in errors.items())
                self.text = base + "\n" + helpful
            else:
                self.text = base
        else:
            self.text = message or ""
            self.code = 0

        fmt = "{0.status} {0.reason} (error code: {1})"
        if len(self.text):
            fmt += ": {2}"

        super().__init__(fmt.format(self.response, self.code, self.text))


class AuthorizationFailure(ClientException):
    """Exception that is raised when authorizing provided token in the Client class fails."""

    pass


class NotFound(HTTPException):
    """HTTPException that is raised when the status code is 404"""

    pass



import typing


__all__ = ("Gist",)


class Gist:
    def __init__(self, data: typing.Dict, client: "Client"):
        self.data = data
        self.client = client
        # Set the data dict's items as attributes
        self._update_attrs(data)

    def _update_attrs(self, data: typing.Dict):
        """Update the Gist object's attributes to the provided data"""

        self.__dict__.update(data)

    async def update(self):
        """Fetch and update the Gist object with the gist"""

        updated_gist_data = await self.client.update_gist(self.id)
        self._update_attrs(updated_gist_data)

    async def edit(self, *, files: typing.Dict, description: str = None):
        """Edit the gist associated with the Gist object, then update the Gist object"""

        edited_gist_data = await self.client.edit_gist(
            self.id, files=files, description=description
        )
        self._update_attrs(edited_gist_data)

    async def delete(self):
        """Delete the gist associated with the Gist object, then delete the Gist object itself"""

        await self.client.delete_gist(self.id)
        del self



from .client import *
from .gist import *
from .exceptions import *

__title__ = "gists"
__author__ = "Witherr"
__license__ = "GNU General Public License v3 (GPLv3)"
__copyright__ = "Copyright 2022-present Witherr"
__version__ = "1.0.3"
